"""
This includes model equation definitions, including error checking functions.

"""
import numpy as np

r = 8.314
bold = '\033[1m'  # for making text bold
unbold = '\033[0m'  # for removing bold text


# Single site langmuir forms
def langmuir1(x, q, b):
    return (q * b * x) / (1 + b * x)


def langmuirlin1(x, q, b):
    return (x / (b * q)) + (1 / (q))


def langmuirlin2(x, q, b):
    return (1 / (b * q)) + (x / q)


# Single site langmuir (temperature dependent form)
def langmuirTD(x, t, q, b0, h):
    b = b0 * np.exp(-h / (r * t))
    return q * ((b * x) / (1 + b * x))


###############################################
# Dual site langmuir forms -  normal and temperature dependent form
def dsl(x, q1, q2, b1, b2):
    site1 = q1 * ((b1 * x) / (1 + b1 * x))
    site2 = q2 * ((b2 * x) / (1 + b2 * x))
    return site1 + site2


def dsltd(x, t, q1, q2, h1, h2, b01, b02):
    b1 = b01 * np.exp(-h1 / (r * t))
    b2 = b02 * np.exp(-h2 / (r * t))
    site1 = q1 * ((b1 * x) / (1 + b1 * x))
    site2 = q2 * ((b2 * x) / (1 + b2 * x))
    return site1 + site2


#################################################
# Modified Dubinin–Radushkevich (MDR) - normal and temperature dependent form
def mdrtd(x, t, n0, n1, a, b, e):
    k = (r * t) / (b * e)
    term1a = (1 - np.exp(- a * x)) * n0
    term1b = np.exp((-k * ((np.log(1 / x)) ** 2)))
    term2 = np.exp(- a * x) * n1 * x
    return term1a * term1b + term2


def mdr(x, n0, n1, a, c):
    term1a = (1 - np.exp(- a * x)) * n0
    term1b = np.exp((-c * ((np.log(1 / x)) ** 2)))
    term2 = np.exp(- a * x) * n1 * x
    return term1a * term1b + term2


################################################

# Henry's law function for the henry region calculation
def henry(x, kh):
    return [i * kh for i in x]


# Calculation for b0 from obtained single langmuir parameters to be used as an initial guess for 'deltaH_langmuir'
def b0_calc(b, h, t):
    return np.exp(np.log(b) + h / (r * t))


###############################################
# BET Extension for type IV and V isotherms (BDDT isotherm model)

# For a maximum (2n-1) layers which can be fit into a capillary:
def bddt1(x, c, n, g, q):
    term1 = (c * x) / (1 - x)
    term2_num = 1 + (((n * g) / 2) - n) * x ** (n - 1) - (n * g - n + 1) * x ** n + (n * g / 2) * x ** (n + 1)
    term2_den = 1 + (c - 1) * x + ((c * g / 2) - c) * x ** n - (c * g / 2) * x ** (n + 1)
    return term1 * (term2_num / term2_den) * q


# For a maximum number of 2n layers:
def bddt2(x, c, n, g, q):
    term1 = (c * x) / (1 - x)
    term2_num = 1 + (((n * g) / 2) - n / 2) * x ** (n - 1) - (n * g + 1) * x ** n + (n * g / 2 + n / 2) * x ** (n + 1)
    term2_den = 1 + (c - 1) * x + ((c * g / 2) - c / 2) * x ** n - (c * g / 2 + c / 2) * x ** (n + 1)
    return term1 * (term2_num / term2_den) * q


# Do and Do isotherm model
def dodo(x, ns, kf, nu, ku, m):
    term1 = (ns * kf * x) / ((1 - x) * (1 + (kf - 1) * x))
    term2 = (nu * ku * x ** m) / (1 + ku * x ** m)
    return term1 + term2

# Guggenheim-Anderson-De Boer (GAB) isotherm model
def gab(x, n, ka, ca):
    num = n * ka * ca * x
    den = (1 - ka * x) * (1 + (ca - 1) * ka * x)
    return num / den

# Brunauer Emmett Teller (BET) isotherm model
def bet(x, n, c):
    return (c * x * n) / ((1 - x) * (1 - x + c * x))


#######################################################
# Sips isotherm model
def sips(x, q, b, n):
    return q * (((b * x) ** (1 / n)) / (1 + (b * x) ** (1 / n)))

# Toth isotherm model
def toth(x, q, b, t):
    return q * ((b * x) / ((1 + (b * x) ** t) ** (1 / t)))

# Toth temperature dependent model
def tothTD(x, temp, q, b0, h, t):
    b = b0 * np.exp(-h / (r * temp))
    return q * ((b * x) / ((1 + (b * x) ** t) ** (1 / t)))

########################################################
# R squared calculation
def r2(x, y, f, c):
    residuals = y - f(x, *c)
    ss_res = np.sum(residuals ** 2)
    ss_tot = np.sum((y - np.mean(y)) ** 2)
    result = 1 - (ss_res / ss_tot)
    if result <= 0:
        # Returns zero if the result is less than zero
        return 0.00000
    else:
        return result


# R squared calculation for henry model
def r2hen(pressures, uptake, f, hen):
    residuals = uptake - f(pressures, hen)
    ss_res = np.sum(residuals ** 2)
    ss_tot = np.sum((uptake - np.mean(uptake)) ** 2)
    result = 1 - (ss_res / ss_tot)
    if result <= 0:
        return 0.00000
    else:
        return result

# Mean squared error calculation
def mse(x, y, f, c):
    y_predicted = f(x, *c)
    return np.square(np.subtract(y, y_predicted)).mean()



import numpy as np
import pandas as pd
import logging as logger

from src.pyIsoFit.core.model_equations import r2hen, henry, bold, unbold
from IPython.display import display


def henry_approx(df, key_pressures, key_uptakes, display_hen=False, tol=0.999, compname="---", henry_off=False):
    """
        Henry approximation function used to estimate henry constants and the henry regime.

    This function works by first adding a (0,0) co-ordinate to each dataset as this is the behaviour exhibited in the
    henry regime. Next, for each temperature dataset, starting with a line made up of three of the first datapoints (
    including the 0,0 point). A gradient (Henry constant) and R squared value of this line is then found through
    linear regression and the package stores it in a variable. It is important to note that the equation that is used
    for linear regression is the Henry isotherm model: q = KH*p.

    This procedure is iterated, with each run adding an additional data point to the line, with the R squared value
    calculated after each fitting. This loop is repeated for the entire dataset. The function stores every calculated
    R2 value as a list, after which the lowest R squared value within this list that is above the minimum tolerance
    value is chosen.

    value. If there are no henry regime lines with an R2 greater than the tolerance value, the henry regime dataset
    corresponding to the maximum R2 value is chosen. The minimum number of datapoints that can be used for the Henry
    regime determination is 3.

    :param df: pd.DataFrame
                Pure-component isotherm data as a pandas dataframe - must be uptake in mmol/g and pressure in bar or
                equivalent. If datasets at different temperatures are required for fitting, the user must specify
                them in the same dataframe.

    :param key_pressures: list[str]
                List of unique column key(s) which correspond to each dataset's pressure values within the
                dataframe. Can input any number of keys corresponding to any number of datasets in the dataframe.
                If multiple dataframes are specified, make sure keys are identical across each dataframe for each
                temperature. Must be inputted in the same order as key_uptakes and temps.

    :param key_uptakes: list[str]
                List of unique column key(s) which correspond to each dataset's uptake values within the
                dataframe. Can input any number of keys corresponding to any number of datasets in the dataframe.
                If multiple dataframes are specified, make sure keys are identical across each dataframe for each
                temperature. Must be inputted in the same order as key_pressures and temps.

    :param display_hen: bool
                Input whether to show the henry regime of the datasets approximated by the package. This is False by
                default.

    :param tol: float or list[float]
                The henry region approximation function calculates the henry region by finding a line with the highest
                R squared value in the low pressure region of the dataset. This is done with a default R squared
                tolerance value (set to 0.999).

                For example, if a float is inputted (a different henry tolerance) this will be the henry tolerance value
                used by the function. i.e if 0.98 is inputted the henry regime will be across a large pressure range
                due to the low tolerance for the R squared value of the henry model fitting.

                This function also supports inputting the henry regimes manually. For this, input each henry regime for
                each dataset as a list i.e [1.2, 2.1, ... ]

    :param compname: str or list[str]
                Name of pure component(s) for results formatting.

    :param henry_off : bool
                Input whether to turn off the henry regime fitting constraint when using the standardised fitting
                constraint to langmuir or dsl - this is usually done when fitting experimental data which has a messy
                low pressure region. Default is False.

    :return: tuple
                Returns list of henry constant for each dataset [0], resulting DataFrame object [1], the x and y
                co-ordinates in the form of a dictionary for each dataset [2]
    """
    if henry_off:
        return None, None, None
    # This section finds the henry region of the datasets
    x = []
    y = []
    for i in range(len(key_pressures)):
        # Reads data from dataframe with respect to keys and adds a 0,0 point
        xi = np.insert(np.array(df[key_pressures[i]].values), 0, 0)
        yi = np.insert(np.array(df[key_uptakes[i]].values), 0, 0)
        x.append(xi)
        y.append(yi)
        del xi
        del yi

    henry_constants = []
    henry_limits = []
    henry_rsq = []
    henry_len = []
    err_lowpoints = []
    err_highrsq = []
    henidx_lst = []
    i = 0

    # Iterate for all datasets at each temperature
    for dataset in y:
        rsq = 1
        x_i = x[i]
        x_henry = [x_i[0], x_i[1]]  # Starting with a minimum of three datapoints
        j = 2
        # Create inner empty lists
        # rsq_ilst stores all calculated r squared values within the inner loops
        # hen_ilst stores all calculated henry constants within the inner loops
        rsq_ilst = []
        hen_ilst = []
        # This loop adds data points while the points correspond to a henry fit with an R^2 of above 0.9999
        if type(tol) == list:
            # Procedure when a list of henry regime limits are passed
            # Iterate over each datapoint within dataset for a given temperature
            while x_i[j] < tol[i] and j < len(x_i) - 3:
                # Add datapoint
                x_henry.append(x_i[j])
                # Create y axis - same length as x axis
                y_henry = dataset[:len(x_henry)]
                # Calculate henry constant
                hen = y_henry[-1] / x_henry[-1]
                # Find r squared value of fit
                rsq = round(r2hen(x_henry, y_henry, henry, hen), 5)
                # Append results to inner lists
                rsq_ilst.append(rsq)
                hen_ilst.append(hen)
                j += 1

            rsqidx = len(hen_ilst) - 1

        else:
            # Procedure when an r squared tolerance is passed
            while rsq > 0 and j < len(x_i):
                # Within the loop - same procedure as in the above condition
                x_henry.append(x_i[j])
                y_henry = dataset[:len(x_henry)]

                hen = y_henry[-1] / x_henry[-1]
                rsq = round(r2hen(x_henry, y_henry, henry, hen), 5)  # r squared calc.
                rsq_ilst.append(rsq)
                hen_ilst.append(hen)
                j += 1

            # Create empty lists
            # abtol stores list of r squared values above the tolerance value
            # itol stores a list of the index values within rsq_ilst
            abtol = []
            itol = []
            i2 = 0
            for rsq in rsq_ilst:
                if rsq > tol:
                    abtol.append(rsq)
                    itol.append(i2)
                i2 += 1

            # Check for an empty list of r squared values above the tolerance value - this can happen if theres a poor
            # Henry regime fitting or if an r squared tolerance is given which is too high
            if abtol == []:
                # In this case the line with the highest r squared value is chosen
                maxrsq = max(rsq_ilst)
                rsqidx = rsq_ilst.index(maxrsq)
                err_highrsq.append(str(i))

            else:
                # Choose the minimum value from the above tolerance r squared list
                rsqfin = min(abtol)
                minidx = abtol.index(rsqfin)
                rsqidx = itol[minidx]

        # +2 to compensate for the initial datapoints
        henidx = rsqidx + 2
        henidx_lst.append(henidx)

        # try:
        henry_len.append(henidx+1)
        # Saving Henry region parameters to later display
        henry_constants.append(hen_ilst[rsqidx])
        henry_limits.append(x_henry[henidx])
        henry_rsq.append(rsq_ilst[rsqidx])
            # sometimes data may not have a good henry region fit, which could abort the above while loop after the
            # first iteration. This piece of code warns the user of this
        # except IndexError:
        #     logger.error("ERROR - Increase henry region value of index " + str(i))

        # Record the index of the dataset where a henry regime is made up of less than 4 points
        if henidx < 3:
            err_lowpoints.append(str(i))

        i += 1

    # Create resulting henry regime datasets and make dictionary
    x_result = [x[i][:henidx_lst[i]+1] for i in range(len(x))]
    y_result = [y[i][:henidx_lst[i]+1] for i in range(len(y))]
    xy_dict = {"x": x_result, "y": y_result}

    # Creating dataframe for henry constants
    df_henry = pd.DataFrame(list(zip(henry_constants, henry_limits, henry_len, henry_rsq)),
                            columns=['Henry constant (mmol/(bar.g))',
                                     'Upper limit (bar)', 'datapoints', 'R squared'])

    if display_hen:
        print(bold + '\nHenry regime for component ' + compname + ':' + unbold)
        display(pd.DataFrame(df_henry))

    # Print a warning if a dataset contains less than 4 points within the henry regime
    if err_lowpoints and display_hen:
        logger.warning(
            unbold + f'WARNING for Dataset {", ".join(err_lowpoints)}: Datasets were found to be made up of less '
                     f'than 4 points.\n '
                     f'Henry region tolerance may be entered as a kwarg - either as a float (default = 0.999) '
                     f'or a list of floats specifying the upper limit for the henry regime for each dataset.\n')

    if err_highrsq and display_hen:
        logger.warning(
            f"WARNING for Dataset {', '.join(err_highrsq)}: Could not find a line with an r squared of above {tol}. "
            f"Creating a line with the highest r squared value. Consider adding henry regime manually by passing"
            f"hen_tol as a list of floats corresponding to the henry regimes of each dataset i.e hen_tol = [0.1, ...,]"
            f"\n")

    return henry_constants, df_henry, xy_dict


"""
Main module for all of the features that pyIsoFit-master has.
"""
import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
from src.pyIsoFit.core.model_fit_def import get_guess_params
from src.pyIsoFit.core.model_dicts import _MODEL_FUNCTIONS, _MODEL_PARAM_LISTS
from src.pyIsoFit.ext_models.ext_dsl import ext_dsl
from src.pyIsoFit.models.dsl import dsl_fit
from src.pyIsoFit.models.generic import generic_fit
from src.pyIsoFit.core.utility_functions import get_sorted_results, get_xy, heat_calc, plot_settings, \
    get_subplot_size, colours, save_func
from src.pyIsoFit.core.model_equations import mse, henry
from src.pyIsoFit.core.exceptions import ParameterError, SaveError
import logging
from IPython.display import display

from src.pyIsoFit.models.henry import henry_approx

logger = logging.getLogger('pyIsoFit-master')

_MODELS = ['mdr', 'mdr td', 'langmuir', 'langmuir linear 1', 'langmuir linear 2', 'langmuir td',
           'dsl nc', 'dsl', 'gab', 'sips', 'toth', 'toth td', 'bddt', 'bddt 2n', 'bddt 2n-1',
           'dodo', 'bet', 'henry']

_does_something = ['langmuir', 'linear langmuir 1', 'linear langmuir 2', 'langmuir td', 'bddt', 'bddt 2n-1', 'bddt 2n']


class IsothermFit:
    """
    A class used to characterise pure-component isotherm data with an isotherm model and
    utilise the results to predict co-adsorption using extended models.

    The class is split up into four methods - fit, plot, save and plot_emod (plot extended model)
    The IsothermFit class is instantiated by passing isotherm data into it.

    """

    def __init__(self,
                 df=None,
                 temps=None,
                 key_pressures=None,
                 key_uptakes=None,
                 model=None,
                 compname=None,
                 temp_units='C',
                 ):
        """
        :param df: pd.DataFrame or list[pd.DataFrame]
                    Pure-component isotherm data as a pandas dataframe - must be uptake in mmol/g and pressure in bar or
                    equivalent. If datasets at different temperatures are required for fitting, the user must specify
                    them in the same dataframe. A list of dataframes may be passed for the dual-site Langmuir isotherm
                    model procedure where parameter results across different components are utilised. Must be inputted
                    in the same order as compname (when passing as a list).

        :param temps: list[float]
                    List of temperatures corresponding to each dataset within the dataframe for results formatting and
                    for calculating heats of adsorption/ binding energies. Must be inputted in the same order as
                    key_pressures and key_uptakes.

        :param key_pressures: list[str]
                    List of unique column key(s) which correspond to each dataset's pressure values within the
                    dataframe. Can input any number of keys corresponding to any number of datasets in the dataframe.
                    If multiple dataframes are specified, make sure keys are identical across each dataframe for each
                    temperature. Must be inputted in the same order as key_uptakes and temps.

        :param key_uptakes: list[str]
                    List of unique column key(s) which correspond to each dataset's uptake values within the
                    dataframe. Can input any number of keys corresponding to any number of datasets in the dataframe.
                    If multiple dataframes are specified, make sure keys are identical across each dataframe for each
                    temperature. Must be inputted in the same order as key_pressures and temps.

        :param model: str
                    Model to be fit to dataset(s).

        :param compname: str or list[str], optional
                    Name of pure component(s) for results formatting. If None is passed, self.compname is instantiated
                    as anarbitrary letter or a list of arbitrary letters corresponding to each component. Must be
                    inputted in the same order as compname (when passing as a list).

        :param temp_units: str, Optional
                    Units of temperature input (temps). Default is degrees C. Can accept Kelvin, 'K'.

        """

        # Checks

        if df is None:
            raise ParameterError("Input Pandas Dataframe with pure-component isotherm data for fitting")

        if temps is None:
            raise ParameterError("Input temperature corresponding to each pure-component isotherm dataset within the "
                                 " dataframe for fitting")

        if key_pressures is None:
            raise ParameterError("Input list of unique column key(s) which correspond to each dataset's pressure "
                                 "values within the Dataframe")

        if key_uptakes is None:
            raise ParameterError("Input list of unique column key(s) which correspond to each dataset's uptake "
                                 "values within the Dataframe")

        if model.lower() is None:
            raise ParameterError("Enter a model as a parameter")

        if model.lower() not in _MODELS:
            raise ParameterError("Enter a valid model - List of supported models:\n " + str(_MODELS))

        len_check = [len(key_uptakes), len(key_pressures), len(temps)]
        if len(temps) != sum(len_check) / len(len_check):
            raise ParameterError("Lengths of key_uptakes, key_pressures or temps do not match. "
                                 "Check that the length of each list is the same, corresponding to each dataset")

        if type(df) is list and model.lower() != "dsl":
            raise ParameterError("Enter one dataframe, not a list of dataframes")

        if compname is None and type(compname) is not list:
            self.compname = 'A'
            logger.info('No component name passed - giving component an arbitrary name.')

        elif compname is None and type(df) is list:
            letters = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K']
            self.compname = [letters[i] for i in range(len(compname))]
            logger.info('No component names passed - giving components arbitrary names.')
            del letters

        logger.info('Checks successfully passed')

        self.df = df
        self.temps = temps
        self.temp_units = temp_units

        # Calculate temps in K if C is provided
        if self.temp_units == 'K':
            self.temps = temps
        else:
            self.temps = [t + 273 for t in temps]

        self.compname = compname
        self.key_pressures = key_pressures
        self.key_uptakes = key_uptakes
        self.model = model.lower()
        self.input_model = model

        # Internal parameters that are passed between methods
        self.x = []
        self.y = []
        self.params = []
        self.df_result = None
        self.emod_input = {}
        self.henry_params = []
        self.rel_pres = False

    def info_params(self):
        """
        Prints information about the model to be fit
        (WIP)

        """
        print(f'Parameters for the {self.model} model:')
        print(_MODEL_PARAM_LISTS[self.model])

    def fit(self,
            cond=False,
            meth='leastsq',
            show_hen=False,
            hen_tol=0.999,
            rel_pres=False,
            henry_off=False,
            guess=None,
            cust_bounds=None,
            fit_report=False,
            weights=None,
            dsl_comp_a=None
            ):

        """
        Plotting method for the FitIsotherm class.
        Fits model to data using Non-Linear Least-Squares Minimization.
        This method is a generic fitting method for all models included in this package using the lmfit
        Parameters and Models class.

        Parameters
        ----------

        :param cond : bool
                Input whether to add standardised fitting constraints to fitting procedure. These are different
                for each fitting. Currently only works for Langmuir, Langmuir td, DSL, BDDT. Default is False

        :param meth : str
                Input the fitting algorithm which lmfit uses to fit curves. Default is 'leastsq' however lmfit includes
                many fitting algorithms which can be inputted (https://lmfit.github.io/lmfit-py/fitting.html).

        :param show_hen : bool
                Input whether to show the henry regime of the datasets approximated by the package. This is False by
                default.

        :param hen_tol : float or list[float]
                The henry region approximation function calculates the henry region by finding a line with the highest
                R squared value in the low pressure region of the dataset. This is done with a default R squared
                tolerance value (set to 0.999).

                For example, if a float is inputted (a different henry tolerance) this will be the henry tolerance value
                used by the function. i.e if 0.98 is inputted the henry regime will be across a large pressure range
                due to the low tolerance for the R squared value of the henry model fitting.

                This function also supports inputting the henry regimes manually. For this, input each henry regime for
                each dataset as a list i.e [1.2, 2.1, ... ]

        :param rel_pres : bool
                Input whether to fit the x axis data to relative pressure instead of absolute. Default is False

        :param henry_off : bool
                Input whether to turn off the henry regime fitting constraint when using the standardised fitting
                constraint to langmuir or dsl - this is usually done when fitting experimental data which has a messy
                low pressure region. Default is False.

        :param guess : dict
                Input custom guess values to override the default guess values. This must be inputted as a dictionary
                with the keys corresponding to the parameter string and the value corresponding to the list of guess
                values corresponding to each dataset.
                i.e for Langmuir: guess = {'q': [5, 5, 6], 'b':[100, 1000, 2000]}

        :param cust_bounds : dict
                Input custom bounds for the fitting. These are hard constraints and lmfit will fit only within these
                minimum and maximum values. Input these as a dictionary with the keys corresponding to the parameter
                string and the value corresponding to the list of tuples which include bounds for each dataset in the
                format (min, max).
                i.e for Langmuir: cust_bounds = {'q': [(4,6), (4, None), (5,10)], ... ect.}

        :param fit_report : bool
                Display a fitting report generated by lmfit for each dataset. Default is False

        :param weights : list[list[float]]
                Weights for fitting

        :param dsl_comp_a : str
                Manually input which component is the most adsorbed component (compA) for the dsl constrained
                fitting procedure.

        :return Returns a dictionary of fitting results


        Note:
        ---------
        Because the dsl constrained fitting procedure fits a list of dataframes, the generic fitting method is not
        used when 'dsl' is inputted with the fitting condition as true and the method returns the result from the
        dsl_fit function. This is because the dsl_fit function carries out its' own initial guess calculations and
        henry regime estimations. The user may interact with this model in the same way as with the rest, however guess
        must be inputted as a list of dictionaries (just as with the list of DataFrames and component names).
        Custom bounds cannot yet be inputted into this model as this is a WIP.
        """
        # When model is dsl and the constrained fitting condtion is true, the generic fitting method is ignored
        # and instead everything within this condition is run
        if self.model == "dsl" and cond is True:
            logger.info('DSL Fitting procedure commenced')

            # Input checks - dsl procedure requires lists of dataframes, components and guesses to function
            if type(self.df) is not list:
                self.df = [self.df]

            if type(self.compname) is not list:
                self.compname = [self.compname]

            if type(guess) is not list and guess is not None:
                guess = [guess]

            # Calling the DSL fitting function - see DSL function docstring for more info
            try:
                dsl_result = dsl_fit(self.df, self.key_pressures, self.key_uptakes,
                                     self.temps, self.compname, meth, guess, hen_tol, show_hen, henry_off, dsl_comp_a)
            except ValueError:
                # Often the fitting algorithm is the source of the error
                logger.critical('The model function generated NaN values and the fit aborted!'
                                'Please check your model function and/or set boundaries on parameters where applicable'
                                'In these cases try changing lmfit fitting method (default is "leastsq") to "tnc"')

            df_dict, results_dict, df_res_dict, params_dict = dsl_result

            # Allocating x and y axis co-ordinates to class variables
            # Produces a list of lists - outer corresponds to component, inner corresponds to temperatures
            self.params = results_dict
            for i in range(len(self.compname)):
                x_i, y_i = df_dict[self.compname[i]]
                self.x.append(x_i)
                self.y.append(y_i)

            # Allocating the resulting dataframes for use within the .save() method
            self.df_result = df_res_dict

            # Allocating the parameter result dictionary for use within the .plot_emod() method
            self.emod_input = params_dict

            return df_res_dict
        logger.info('Generic fitting procedure commenced')

        # cond=True only works for certain models - This notifies the user of this
        if self.model not in _does_something and cond is not False:
            logger.warning(f"WARNING You have set cond={cond} but cond for the model '{self.model}' does nothing.\n")

        # Calculating henry constants and displaying results (see function docstring)
        if self.model == "henry":
            show_hen = True

        henry_params = henry_approx(self.df, self.key_pressures, self.key_uptakes, show_hen, hen_tol,
                                    self.compname, henry_off)

        henry_constants = henry_params[0]

        # When model is henry, only the henry approximation function is run and the rest is ignored
        if self.model == "henry":
            logger.info('Henry model fitting only chosen')
            self.henry_params = henry_params
            self.df_result = henry_params[1]
            return None

        # ! Dictionary of parameters as a starting point for data fitting
        guess = get_guess_params(self.model, self.df, self.key_uptakes, self.key_pressures)
        logger.info('Guess values successfully obtained')

        # Override defaults if user provides param_guess dictionary
        if guess is not None:
            for param, guess_val in guess.items():
                if param not in list(guess.keys()):
                    raise ParameterError("%s is not a valid parameter"
                                         " in the %s model." % (param, self.model))
                guess[param] = guess_val
            logger.info('Guess values overridden with custom guess values')

        # MDR works best when dealing with relative pressures - This forces it on
        if "mdr" in self.model:
            logger.info('MDR chosen so relative pressure toggle force set')
            rel_pres = True
        self.rel_pres = rel_pres

        # Extracts the x and y parameter lists from the dataframe
        self.x, self.y = get_xy(self.df, self.key_pressures, self.key_uptakes, self.model, rel_pres)
        logger.info('x and y parameters successfully obtained')

        if weights is None:
            logger.info('No weights inputted - setting weights to x')
            weights = self.x

        # Everything except the model function is the same for the bddt fitting procedure
        # This reduces extra entries within the model_dicts.py file by standardising the input
        if self.model == "bddt 2n" or self.model == "bddt 2n-1" or self.model == "bddt":
            self.model = "bddt"

        # ----------Generic fitting procedure--------------

        # Allocate list of lmfit result objects to class and create values_dict variable
        self.params, values_dict = generic_fit(self.model, weights, self.y, guess, self.temps, cond, meth, cust_bounds,
                                               fit_report, henry_constants, henry_off)
        logger.info('Generic fit completed successfully')

        # Sorts the values_dict variable to a dictionary which can be used to create a dataframe result and creates a
        # list of result parameters to feed into the error checking functions
        final_results_dict, c_list = get_sorted_results(values_dict, self.model, self.temps)
        logger.info('Results sorted successfully')

        # Find the r squared and mean squared error (mse) values
        # r_sq = [r2(self.x[i], self.y[i], _MODEL_FUNCTIONS[self.model], c_list[i]) for i in range(len(self.x))]
        se = [mse(self.x[i], self.y[i], _MODEL_FUNCTIONS[self.model], c_list[i]) for i in range(len(self.x))]
        logger.info('Mean squared error calculated successfully')

        # Add r squared and mse results to final results dictionary
        # final_results_dict['R squared'] = r_sq
        final_results_dict['MSE'] = se

        # Create dataframe with fitting results
        df_result = pd.DataFrame.from_dict(final_results_dict)
        pd.set_option('display.max_columns', None)

        # Print fitting results
        print(f'\n\n\n---- Component {self.compname} fitting results -----')
        display(df_result)

        # Save dataframes for model results and henry regime to class for use within the .save() function
        self.df_result = df_result

        if len(self.temps) >= 3:
            heat_calc(self.model, self.temps, final_results_dict, self.x)

    def plot(self, logplot=(False, False)):
        """
        Plotting method for the FitIsotherm class.
        There are three plotting procedures:
         - For more than one component
         - For henry plots (This requires plotting on individual subplots)
         - The generic plotting procedure for any other model with one component (most models use this)

        :param logplot: tuple(bool)
            Whether to have an x and y log axis. Default is off for both x and y axis i.e (False, False) in the order
            (x, y)

        """
        np.linspace(0, 10, 301)

        fit_label = "{temps} K Fit"
        data_label = "Data at {temps} K"

        # Plotting for more than one component
        if type(self.df) is list:
            for i in range(len(self.df)):
                # Get plot settings
                plot_settings(logplot)

                comp_x_params = self.params[self.compname[i]]
                plt.title(self.compname[i])
                for j in range(len(self.key_pressures)):
                    plt.plot(self.x[i][j], comp_x_params[j].best_fit, '-', color=colours[j],
                             label=fit_label.format(temps=self.temps[j]))
                    plt.plot(self.x[i][j], self.y[i][j], 'ko', color='0.75',
                             label=data_label.format(temps=self.temps[j]))

        # Plotting for henry model
        elif self.model == "henry":
            henry_const = self.henry_params[0]
            xy_dict = self.henry_params[2]
            x_hen = xy_dict['x']
            y_hen = xy_dict['y']
            plot_settings(logplot)
            lenx = len(self.x)
            for i in range(len(x_hen)):
                y_henfit = henry(x_hen[i], henry_const[i])
                subplot_size = get_subplot_size(lenx, i)

                plt.subplot(subplot_size[0], subplot_size[1], subplot_size[2])
                plt.subplots_adjust(wspace=0.3, hspace=0.3)

                plt.title('Henry regime at ' + str(self.temps[i]) + ' K')
                plt.plot(x_hen[i], y_henfit, '-', color=colours[i],
                         label=fit_label.format(temps=self.temps[i]))
                plt.plot(x_hen[i], y_hen[i], 'ko', color='0.75',
                         label=data_label.format(temps=self.temps[i]))
                plt.legend()

        # Plotting for any other model
        else:
            plot_settings(logplot, self.input_model, self.rel_pres)

            for i in range(len(self.key_pressures)):
                plt.plot(self.x[i], self.params[i].best_fit, '-', color=colours[i],
                         label=fit_label.format(temps=self.temps[i]))
                plt.plot(self.x[i], self.y[i], 'ko', color='0.75',
                         label=data_label.format(temps=self.temps[i]))
        plt.legend()
        plt.show()

    def save(self, directory=None, filestring=None, filetype='.csv'):
        """
        Saves the model fitting result and henry region fitting result dataframes to directory as a .csv or .json
        file.

        :param directory:
            Full destination directory must be inputted for the user to save a file

        :param filestring: list[str] or str
            This is a list of strings corresponding to the file names, first position is fit result, second is
            henry result. Inputting this as a string for the fit result only will also work.

        :param filetype: str
            .csv or .json for saving

        """
        # Directory input check
        if directory is None:
            raise SaveError("\n\nPlease enter full directory for saving file separated by double dashes i.e "
                            "C:\\Users\\User\\pyIsoFit-master\\fittingresults\\")

        # File name input check
        if filestring is None:
            filestring = 'fit_result'

        # Creates file for each component when there are multiple components
        if type(self.df_result) is dict:
            for comp in self.df_result:
                save_func(directory, filestring, filetype, self.df_result[comp], comp)
        else:
            save_func(directory, filestring, filetype, self.df_result)

    def plot_emod(self, yfracs, ext_model='extended dsl', logplot=(False, False)):
        """
        Predicts co-adsorption isotherm data and plots it.

        :param yfracs: list[float]
            List of component mole fractions within the gas mixture

        :param ext_model: str
            Extended model for the method to predict co-adsorption with. Currently extended DSL is the only
            model included

        :param logplot: bool
            Whether to have an x and y log axis.

        :return:
            Returns a dictionary of co-adsorption uptakes for each component
        """

        # Checks
        if len(self.compname) < 2:
            raise ParameterError("Enter 2 components or more to use extended models")

        if self.model != "dsl":
            raise ParameterError("""This isotherm model is not supported for extended models. Currently supported
             models are:
            - DSL """)

        # Only extended dsl is currently supported so no need for if else statements
        q_dict = ext_dsl(self.emod_input, self.temps, self.x, self.compname, yfracs)

        # Plot results ext_dsl function
        for i in range(len(self.compname)):
            plot_settings(logplot)
            q = q_dict[self.compname[i]]
            plt.title(f'Co-adsorption isotherm for component {self.compname[i]} at mol frac of {yfracs[i]}')
            for j in range(len(self.temps)):
                plt.plot(self.x[i][j], q[j], '--', color=colours[j],
                         label="{temps} K Fit".format(temps=self.temps[j]))
            plt.legend()
            plt.show()
        return q_dict



import pandas as pd
import numpy as np
from scipy import stats
from lmfit import Model, Parameters

from src.pyIsoFit.core.model_equations import langmuir1, dsl, dsltd, langmuirTD, r, bold, unbold, mse
from src.pyIsoFit.core.model_fit_def import get_guess_params
from src.pyIsoFit.models.henry import henry_approx
from IPython.display import display


def dsl_fit(df_list: list,
            key_pressures: list,
            key_uptakes: list,
            temps: list,
            compnames: list,
            meth: str,
            guess: list,
            hentol: float,
            show_hen: bool,
            henry_off: bool,
            dsl_comp_a: str
            ):
    """
    Thermodynamically consistent fitting procedure written with accordance to thermodynamic consistency principles.
    These are similar procedures to ones described by Farmahini et al. (2018) (doi.org/10.1021/acs.iecr.8b03065).

    Comprised of three inner functions, corresponding to each step in the procedure, which are called sequentially
    for each component.

    The outer function extracts data from Pandas Dataframes, sorts them, finds initial guess values, finds henry
    constants and calls the inner functions to carry out fitting.


    :param df_list: list[DataFrame]
            List of Pandas Dataframes corresponding to each component.

    :param compnames: list[str]
            List of component names corresponding to each component.

    :param guess: list[dict]
            List of guess dictionaries corresponding to each component.

    :return: tuple of:
        [0] : Dictionary of x and y co-ordinates corresponding to each component for all temperatures in the form of:
            {'component n' : {'x': [list[x1], list[x2],..., list[xn]],
                            {'y': [list[y1], list[y2],..., list[yn]]},
            'component n+1' : {...}
            }

        [1] : Dictionary of lmfit ModelResult objects corresponding to each component for all temperatures in the
                form of:
            {'component n' : list[ModelResult], 'component n+1': ... }

        [2] : Dictionary of Pandas Dataframe objects corresponding to each component in the form of:
            {'component n' : Dataframe, 'component n+1' : ...}

        [3] : Dictionary of best fitting parameters corresponding to each component in the form of:
            {'component n' : {'q1' : list[pars],
                              'q2' : list[pars],
                              ...
                              },
            'component n+1' : {...}
            }


        xy_dict, results_dict, df_res_dict, params_dict
    """
    def step1(x, y, meth, guess, henry_constants):
        """
        The fitting procedure is started from the lowest temperature dataset as experimentally this leads to the
        determination to the closest value to the actual saturated adsorption capacity of the material.

        The model must also reduce to Henry’s law at a limit of zero
        pressure, and so the following constraint must be set in the fitting procedure:

        {KH = (q1 * b1) + (q2 * b2)}


        :return:
            Returns initial qsat values to be fixed for the remaining datasets.

        Note - This function is first used for finding the most adsorbed component from a list of dataframes
        """

        gmod = Model(dsl)
        pars1 = Parameters()

        pars1.add('q1', value=guess['q1'][0], min=0)
        pars1.add('q2', value=guess['q2'][0], min=0)
        pars1.add('b2', value=guess['b2'][0], min=0)
        if henry_off:
            pars1.add('b1', value=guess['b1'][0], min=0)
        else:
            pars1.add('delta', value=henry_constants[0], vary=False)
            pars1.add('b1', expr='(delta-q2*b2)/q1')  # KH = b1*q1 + b2*q2

        result1 = gmod.fit(y[0], pars1, x=x[0], method=meth)

        return result1.values['q1'], result1.values['q2']

        # This ends the function within the inner class and returns the qmax values to
        # be used in step 2

    def step2(x, y, meth, guess, henry_constants, temps, step1_pars, comp2=False):
        """
        The same procedure as in step 1 is followed for the remaining isotherm datasets with the same constraints as
        described in step 1. q1 and q2 are fixed at that of the first temperature dataset.
        Fitting results in q1, q2, b1 and b2 being found, after which Van't Hoff plots are made to find b01, b02, h1
        and h2.

        :return:
            Returns calculated b01, b02, h1 and h2 for use in step 3
        """
        q1fix, q2fix = step1_pars

        if comp2:
            # Depending on whether the package is fitting for the most adsorbed componennt or the
            # remaining components, the fitting procedure is different. In the methodology used,
            # The most adsorbed component is fitted to DSL but the remaining components are
            # fit to essentially the single langmuir isotherm
            gmod = Model(langmuir1)
            qtot = q1fix + q2fix

            c_list = []
            for i in range(len(x)):
                pars = Parameters()
                pars.add('q', value=qtot, min=qtot, max=qtot + 0.001)
                if henry_off:
                    pars.add('b2', value=guess['b2'][i])
                else:
                    pars.add('delta', value=henry_constants[i], vary=False)
                    pars.add('b', expr='delta/q')

                results = gmod.fit(y[i], pars, x=x[i], method=meth)
                c = [q1fix, q2fix, results.values['b'], results.values['b']]
                c_list.append(c)

                del results
                del pars

        else:
            gmod = Model(dsl)
            c_list = []
            for i in range(len(x)):
                pars = Parameters()
                pars.add('q1', value=q1fix, min=q1fix, max=q1fix + 0.001)
                pars.add('q2', value=q2fix, min=q2fix, max=q2fix + 0.001)
                pars.add('b2', value=guess['b2'][i], min=0)
                if henry_off:
                    pars.add('b1', value=guess['b1'][i])
                else:
                    pars.add('delta', value=henry_constants[i], vary=False)
                    pars.add('b1', expr='(delta-q2*b2)/q1', min=0)  # KH = b*q

                results = gmod.fit(y[i], pars, x=x[i], method=meth)
                c = [results.values['q1'], results.values['q2'],
                     results.values['b1'], results.values['b2']]
                c_list.append(c)

                del results
                del pars

        # allocating variables
        qsat1 = c_list[0][0]
        qsat2 = c_list[0][1]
        b1 = [param[2] for param in c_list]
        b2 = [param[3] for param in c_list]

        # Finding heat of adsorption for both sites
        T = np.array([1 / temp for temp in temps])
        ln_b1 = np.array([np.log(i) for i in b1])
        ln_b2 = np.array([np.log(i) for i in b2])
        mH1, bH1, rH1, pH1, sterrH1 = stats.linregress(T, ln_b1)
        mH2, bH2, rH2, pH2, sterrH2 = stats.linregress(T, ln_b2)

        h = [-mH1 * r, -mH2 * r]
        b0 = [np.exp(bH1), np.exp(bH2)]

        return qsat1, qsat2, h, b0
        # The package returns these sets of parameters to be used as initial guesses for the final step

    def step3(x, y, meth, temps, step2_pars, comp2=False):
        """
        The temperature dependent form of the DSL equation is now used for fitting, which is the DSL equation
        with the Van’t Hoff equation substituted for b1 and b2 giving a model with 6 fitting
        parameters; b01, b02, h1, h2, q1, q2. The parameters calculated at the end of step 2 are used as
        initial guess values for this fitting.

        :return: tuple of:
                A list of lmfit results objects [0], the results in the form of a dataframe [1],
                a dictionary of the fitting results
        """
        # unpacking tuples
        q1_in, q2_in, h_in, b0_in = step2_pars

        # We can now fit to the van't Hoff form of DSL with the initial guess values
        # obtained through step 2

        if comp2 == True:
            # Again, we fit to the single langmuir form when the least adsorbed species is used
            gmod = Model(langmuirTD)
            qtot = q1_in + q2_in
            c_list = []
            results_lst = []
            for i in range(len(x)):
                pars = Parameters()
                pars.add('t', value=temps[i], vary=False)
                pars.add('q', value=qtot, min=qtot, max=qtot + 0.001)
                pars.add('h', value=h_in[0])
                pars.add('b0', value=b0_in[0], min=0)

                results = gmod.fit(y[i], pars, x=x[i], method=meth)
                results_lst.append(results)

                c_list.append([results.values['t'], q1_in, q2_in, results.values['h'],
                               results.values['h'], results.values['b0'], results.values['b0']])

                del results
                del pars

        else:
            gmod = Model(dsltd)  # DSL
            c_list = []
            results_lst = []
            for i in range(len(x)):
                pars = Parameters()
                pars.add('t', value=temps[i], vary=False)
                pars.add('q1', value=q1_in, min=q1_in, max=q1_in + 0.001)
                pars.add('q2', value=q2_in, min=q2_in, max=q2_in + 0.001)
                pars.add('h1', value=h_in[0])
                pars.add('h2', value=h_in[1])
                pars.add('b01', value=b0_in[0], min=0)
                pars.add('b02', value=b0_in[1], min=0)

                results = gmod.fit(y[i], pars, x=x[i], method=meth)
                results_lst.append(results)
                c_list.append([results.values['t'], results.values['q1'], results.values['q2'], results.values['h1'],
                               results.values['h2'], results.values['b01'], results.values['b02']])

                del results
                del pars

        print(bold + "\nParameters found..." + unbold)

        t = [param[0] for param in c_list]

        # Allocating variables
        qmax1 = [param[1] for param in c_list]
        qmax2 = [param[2] for param in c_list]
        h1 = [param[3] for param in c_list]
        h2 = [param[4] for param in c_list]
        b01 = [param[5] for param in c_list]
        b02 = [param[6] for param in c_list]

        params_dict = {'q1': qmax1,
                       'q2': qmax2,
                       'h1': h1,
                       'h2': h2,
                       'b01': b01,
                       'b02': b02
                       }

        # Checking mse squared of fits
        se = [mse(x[i], y[i], dsltd, c_list[i]) for i in range(len(x))]

        # Displaying results
        pd.set_option('display.max_columns', None)
        df_result = pd.DataFrame(list(zip(t, qmax1, qmax2, h1, h2, b01, b02, se)),
                                 columns=['Temp(K)', 'qmax1 (mmol/g)',
                                          'qmax2 (mmol/g)', 'h1 (J/mol)', 'h2 (J/mol)', 'b01 (1/bar)', 'b02 (1/bar)',
                                          'mse'])

        display(pd.DataFrame(df_result))

        print(bold + "===============================================================================")
        print(bold + "===============================================================================")

        return results_lst, df_result, params_dict

    # Extracting x and y from datasets for each component
    xy_dict = {}
    df_res_dict = {}
    results_dict = {}
    params_dict = {}
    i = 0
    for df in df_list:
        x = [df[key_pressures[j]].values for j in range(len(key_pressures))]
        y = [df[key_uptakes[j]].values for j in range(len(key_pressures))]
        xy_dict[compnames[i]] = x, y
        del x, y
        i += 1

    # Creating a list of dictionaries of guess values for each component
    if guess is None:
        guess = [get_guess_params("dsl", df_list[i], key_uptakes, key_pressures) for i in range(len(df_list))]

    # Creating a list of henry constants for each component
    henry_const_lst = [henry_approx(df_list[i], key_pressures, key_uptakes, show_hen, hentol, compnames[i])[0] for i in
                       range(len(df_list))]

    # Checking the most adsorbed component - calling function step 1 for all components and checking which gives
    # highest q1 + q2.
    i_high = 0
    qhigh = 0
    if dsl_comp_a is None:
        for i in range(len(compnames)):
            qtest = step1(*xy_dict[compnames[i]], meth, guess[i], henry_const_lst[i])
            qcheck = qtest[0] + qtest[1]
            if qcheck > qhigh:
                # When the calculated qsat(total) is the highest of all components, that qsat is saved along with the
                # index of the component with the highest qsat
                qhigh = qcheck
                i_high = i
        comp_msg = f'{compnames[i_high]} was found to have the highest qsat(total) This will be used as ' \
                   f'component A. '
    else:
        # If the most adsorbed component name is inputted manually,
        i_high = compnames.index(dsl_comp_a)
        qtest = step1(*xy_dict[compnames[i_high]], meth, guess[i_high], henry_const_lst[i_high])
        qhigh = qtest[0] + qtest[1]
        comp_msg = f'You have manually inputted {compnames[i_high]} as the most adsorbed component'

    if len(df_list) > 1:
        print(comp_msg)
        print(f'{compnames[i_high]} shows a qsat(total) of {str(round(qhigh, 1))} mmol/g')
        # Component A refers to the fitting procedure for the most adsorbed component - i.e this component is fit to
        # the DSL isotherm model, while the rest of the components are fit to the single site Langmuir isotherm model


    # Allocates the most adsorbed dataframe to be used in the procedure
    print(bold + "_________________________" + compnames[i_high] + " RESULTS_____________________________")
    print(' ')

    # Calling the inner functions corresponding to the three step fitting procedure
    step1_compA = step1(*xy_dict[compnames[i_high]], meth, guess[i_high], henry_const_lst[i_high])

    step2_compA = step2(*xy_dict[compnames[i_high]], meth, guess[i_high], henry_const_lst[i_high], temps, step1_compA)

    step3_compA = step3(*xy_dict[compnames[i_high]], meth, temps, step2_compA)

    # Allocating the resulting tuple from step 3 to dictionaries to be returned by the outer function
    results_dict[compnames[i_high]] = step3_compA[0]
    df_res_dict[compnames[i_high]] = step3_compA[1]
    params_dict[compnames[i_high]] = step3_compA[2]

    # Iterating the procedure for 'component B' for all other dataframes (except the one with index i_high)
    for i in range(len(df_list)):
        if len(df_list) > 1 and i != i_high:
            print(bold + "_________________________" + compnames[i] + " RESULTS_____________________________")
            print(' ')
            step2_compB = step2(*xy_dict[compnames[i]], meth, guess[i], henry_const_lst[i], temps,
                                step1_compA, True)

            step3_compB = step3(*xy_dict[compnames[i]], meth, temps, step2_compB, True)
            results_dict[compnames[i]] = step3_compB[0]
            df_res_dict[compnames[i]] = step3_compB[1]
            params_dict[compnames[i]] = step3_compB[2]

    return xy_dict, results_dict, df_res_dict, params_dict



"""
Module for functions associated with model definitions.

"""
from src.pyIsoFit.core.utility_functions import bounds_check
from src.pyIsoFit.models.henry import henry_approx


def get_guess_params(model, df, key_uptakes, key_pressures):
    """
    Function for automatic initial guess value calculation based on a langmuir henry regime approximation.

        - The saturation loading is estimated by obtaining the highest uptake value for each dataset.
        - The henry_approx function (utility_functions.py) is used here to calculate the henry constants
        - The langmuir 'b' parameter is calculated from the henry constants based on { KH = b * q }
        - The heat of adsorption guess value is set to -5000 for all datasets as this is difficult to estimate
        based on the datasets given

    :return:
        Returns dictionary of guess values (as lists) for each parameter corresponding to the chosen model

    """
    saturation_loading = [1.1 * df[key_uptakes[i]].max() for i in range(len(key_pressures))]
    henry_lim = henry_approx(df, key_pressures, key_uptakes, False)[0]
    langmuir_b = [kh / qsat for (kh, qsat) in zip(henry_lim, saturation_loading)]
    h_guess = [-5000 for _ in key_pressures]

    if "langmuir" in model and model != "langmuir td":
        return {
            "b": langmuir_b,
            "q": saturation_loading
        }

    if model == "langmuir td":
        return {
            "b0": langmuir_b,
            "q": saturation_loading,
            "h": h_guess
        }

    if model == "dsl" or model == "dsl nc":
        return {
            "q1": [0.5 * q for q in saturation_loading],
            "b1": [0.4 * b for b in langmuir_b],
            "q2": [0.5 * q for q in saturation_loading],
            "b2": [0.6 * b for b in langmuir_b]
        }

    if model == "gab":
        return {
            "n": [0.2 * q for q in saturation_loading],
            "ka": [1.1 * b for b in langmuir_b],
            "ca": [0.1 * b for b in langmuir_b]
        }

    if model == "mdr":
        return {
            "n0": saturation_loading,
            "n1": langmuir_b,
            "a": [0.1 * b for b in langmuir_b],
            "c": [10 * b for b in langmuir_b]
        }
    if model == "sips":
        return {
            "q": saturation_loading,
            "b": langmuir_b,
            "n": [1 for _ in saturation_loading]
        }
    if model == "toth":
        return {
            "q": saturation_loading,
            "b": langmuir_b,
            "t": [0.5 for _ in saturation_loading]
        }
    if "bddt" in model:
        return {
            "c": langmuir_b,
            "n": [10 for _ in saturation_loading],
            "g": [100 for _ in saturation_loading],
            "q": [q * 0.5 for q in saturation_loading]
        }
    if model == "dodo":
        return {
            "ns": saturation_loading,
            "kf": langmuir_b,
            "nu": [b * 10 for b in saturation_loading],
            "ku": langmuir_b,
            "m": [5 for _ in saturation_loading]
        }
    if model == "bet":
        return {
            "n": saturation_loading,
            "c": langmuir_b,
        }
    if model == "toth td":
        return {
            "q": saturation_loading,
            "b0": [b * 0.1 for b in langmuir_b],
            "t": [0.5 for _ in saturation_loading],
            "h": h_guess
        }


def get_fit_tuples(model, guess, temps, i, cond, cust_bounds, henry_constants, henry_off, q_fix):
    """
    Definitions of the fitting procedures for each model in the form of tuples.

    This function has an iterative element (i) as when iterating over a list of datasets for fitting, each fitting
    will have its own unique guess values and bounds.

    These tuples are formed for the .add_many() method in the lmfit Parameters class. The following definition is
    extracted from https://lmfit.github.io/lmfit-py/parameters.html :
        add_many(*parlist) - Add many parameters, using a sequence of tuples.

        Parameters *parlist (sequence of tuple or Parameter) – A sequence of tuples, or a sequence of Parameter
        instances. If it is a sequence of tuples, then each tuple must contain at least a name. The order in each
        tuple must be (name, value, vary, min, max, expr, brute_step).

    :param q_fix: float
        Used to fix the saturation capacity parameter for langmuir.


    :return:
        Returns a tuple of parameters for the .add_many() method in the lmfit Parameters class.
    """
    bounds = bounds_check(model, cust_bounds, len(temps))

    if model == "dsl" and cond is False:
        return ('q1', guess['q1'][i], True, *bounds['q1'][i]), \
               ('q2', guess['q2'][i], True, *bounds['q2'][i]), \
               ('b1', guess['b1'][i], True, *bounds['b1'][i]), \
               ('b2', guess['b2'][i], True, *bounds['b2'][i])

    if "langmuir" in model and model != "langmuir td":
        q = ('q', guess['q'][i], True, *bounds['q'][i])
        b = ('b', guess['b'][i], True, *bounds['b'][i])
        if cond:
            if i != 0:
                q = ('q', q_fix, True, q_fix, q_fix + 0.001)

            if henry_off:
                return q, b
            else:
                delta = ('delta', henry_constants[i], False)
                b = ('b', None, None, None, None, 'delta/q')
                return q, delta, b
        return q, b

    if model == "langmuir td":
        q = ('q', guess['q'][i], True, *bounds['q'][i])
        if cond and i != 0:
            q = ('q', q_fix, True, q_fix, q_fix + 0.001)

        return ('t', temps[i], False), \
               q, \
               ('h', guess['h'][i], True, *bounds['h'][i]), \
               ('b0', guess['b0'][i], True, *bounds['b0'][i])

    if model == "gab":
        return ('n', guess['n'][i], True, *bounds['n'][i]), \
               ('ka', guess['ka'][i], True, *bounds['ka'][i]), \
               ('ca', guess['ca'][i], True, *bounds['ca'][i])

    if model == "mdr":
        n0 = ('n0', guess['n0'][i], True, *bounds['n0'][i])
        n1 = ('n1', guess['n1'][i], True, *bounds['n1'][i])
        a = ('a', guess['a'][i], True, *bounds['a'][i])
        c = ('c', guess['c'][i], True, *bounds['c'][i])
        if cond:
            if i != 0:
                n0 = ('n0', q_fix, True, q_fix, q_fix + 0.001)
            if henry_off:
                return n0, n1, a, c
            else:
                delta = ('delta', henry_constants[i], False)
                n1 = ('n1', None, None, None, None, 'delta/n0')
                return n0, delta, n1, a, c

        return n0, n1, a, c

    if model == "sips":
        return ('q', guess['q'][i], True, *bounds['q'][i]), \
               ('b', guess['b'][i], True, *bounds['b'][i]), \
               ('n', guess['n'][i], True, *bounds['n'][i])

    if model == "toth":
        return ('q', guess['q'][i], True, *bounds['q'][i]), \
               ('b', guess['b'][i], True, *bounds['b'][i]), \
               ('t', guess['t'][i], True, *bounds['t'][i])

    if "bddt" in model:
        if cond is True:
            c_con = ('c', guess['c'][i], True, 0, 1)
        else:
            c_con = ('c', guess['c'][i], True, *bounds['c'][i])
        return c_con, \
               ('n', guess['n'][i], True, *bounds['n'][i]), \
               ('g', guess['g'][i], True, *bounds['g'][i]), \
               ('q', guess['q'][i], True, *bounds['q'][i])

    if model == "dodo":
        return ('ns', guess['ns'][i], True, *bounds['ns'][i]), \
               ('kf', guess['kf'][i], True, *bounds['kf'][i]), \
               ('nu', guess['nu'][i], True, *bounds['nu'][i]), \
               ('ku', guess['ku'][i], True, *bounds['ku'][i]), \
               ('m', guess['m'][i], True, *bounds['m'][i])

    if model == "bet":
        return ('n', guess['n'][i], True, *bounds['n'][i]), \
               ('c', guess['c'][i], True, *bounds['c'][i]),



""" Custom errors thrown by the package"""


class pgError(Exception):
    """Base error raised by pyIsoFit-master"""


class ParameterError(pgError):
    """Raised when one of the parameters is unsuitable"""


class SaveError(pgError):
    """Raised when the .save method is called incorrectly"""


class HenryError(pgError):
    """Raised when there is a problem with the henry regime approximation"""



"""
Testing file (not unit test)
---------------------------
"""

from fitting import IsothermFit
import pandas as pd

# df1 = pd.read_csv('SIFSIX-3-Cu CO2.csv')
# df1 = pd.read_csv('../Datasets for testing/Lewatit CO2.csv')
# df1 = pd.read_csv('../../../tests/Datasets for testing/Computational Data (EPFL) CO2.csv')
# df2 = pd.read_csv('../../../tests/Datasets for testing/Computational Data (EPFL) N2.csv')
#
# # df1 = pd.read_csv('../Datasets for testing/data.csv')
# df_list = [df1, df2]
# compname = 'CO2'
# # temps = [0, 25, 50]
# temps = [10, 40, 100]
# # temps = [0, 20]
# # temps = [25, 45, 55]
#
# # keyUptakes = ['y1', 'y2', 'y3']
# # keyPressures = ['x1', 'x2', 'x3']
#
#
# keyUptakes = ['Uptake (mmol/g)_13X_10 (°C)', 'Uptake (mmol/g)_13X_40 (°C)', 'Uptake (mmol/g)_13X_100 (°C)']
# keyPressures = ['Pressure (bar)', 'Pressure (bar)', 'Pressure (bar)']
#
# # keyUptakes = ['q 298', 'q 318', 'q 328']
# # keyPressures = ['p 298', 'p 318', 'p 328']
#
# # keyUptakes = ['Loading [mmol/g] 25', 'Loading [mmol/g] 50', 'Loading [mmol/g] 70']
# # keyPressures = ['Pressure [bar] 25', 'Pressure [bar] 50', 'Pressure [bar] 70']
#
# # cust_bounds = {
# #     'q': [(4.97, 4.99), (4.81, 4.83), (0, 6)],
# #     'b': [(39.5, 39.7), (0, 50), (0, 1000)]
# # }
#
# langmuir = IsothermFit(df1, temps, keyPressures, keyUptakes, "mdr", compname)
# langmuir.fit(cond=False, show_hen=True)
# langmuir.plot(logplot=True)
# # langmuir.save()
# # langmuir.plot_emod(yfracs=[0.15, 0.85], logplot=False)


df1 = pd.read_csv('../../../tests/Datasets for testing/Lewatit CO2.csv')
compname = 'CO2'
temps = [25, 50, 75, 100]
key_pressures = ['p1', 'p2', 'p3', 'p4']
key_uptakes = ['q1', 'q2', 'q3', 'q4']

test = IsothermFit(df1, temps, key_pressures, key_uptakes, "langmuir linear 2", compname)
test.fit(hen_tol=0.999, show_hen=True)
# test.save(directory="C:\\Users\\domin\\PycharmProjects\\pyIsoFit-master\\fittingresults\\")
test.plot()

# df1 = pd.read_csv('../../../tests/Datasets for testing/CMS.csv')
# compname = 'H2O'
# temps = [20, 50, 60]
# key_pressures = ['p 20', 'p 50', 'p 60']
# key_uptakes = ['q1', 'q2', 'q3']
#
# test = IsothermFit(df1, temps, key_pressures, key_uptakes, "bddt", compname)
# test.info_params()
# test.fit(rel_pres=True)
# test.plot()



"""
This file includes all of the dictionaries which are necessary for fitting.

If a model is to be added using a generic fitting method, it may be added by adding entries to the dictionaries below as
well as the model_fit_def.py file.
"""
from src.pyIsoFit.core.model_equations import langmuir1, langmuirlin1, langmuirlin2, langmuirTD, gab, mdr, dsl, sips, \
    toth, bddt1, bddt2, dodo, bet

# List of parameters for each model used for sorting dictionaries
_MODEL_PARAM_LISTS = {
    'mdr': ['n0', 'n1', 'a', 'c'],
    'langmuir': ['q', 'b'],
    'langmuir linear 1': ['q', 'b'],
    'langmuir linear 2': ['q', 'b'],
    'langmuir td': ['q', 'b0', 'h'],
    'dsl': ['q1', 'q2', 'b1', 'b2'],
    'gab': ['n', 'ka', 'ca'],
    'sips': ['q', 'b', 'n'],
    'toth': ['q', 'b', 't'],
    'bddt': ['c', 'n', 'g', 'q'],
    'dodo': ['ns', 'kf', 'nu', 'ku', 'm'],
    'bet': ['n', 'c']
}

# List of dataframe titles for each model, used for creating final result dataframe
# Input units here beside parameters for clarity when describing parameters
_MODEL_DF_TITLES = {
    'mdr': ['n0', 'n1', 'a', 'c'],
    'langmuir': ['q (mmol/g)', 'b (1/bar)'],
    'langmuir linear 1': ['q (mmol/g)', 'b (1/bar)'],
    'langmuir linear 2': ['q (mmol/g)', 'b (1/bar)'],
    'langmuir td': ['q (mmol/g)', 'b0 (1/bar)', 'h (kJ/mol)'],
    'dsl': ['q1 (mmol/g)', 'q2 (mmol/g)', 'b1 (1/bar)', 'b2 (1/bar)'],
    'gab': ['n (mmol/g)', 'ka (H2O activity coeff.)', 'ca (GAB const.)'],
    'sips': ['q (mmol/g)', 'b (1/bar)', 'n (heterogeneity parameter)'],
    'toth': ['q (mmol/g)', 'b (1/bar)', 't (heterogeneity parameter)'],
    'bddt': ['c (BET const.)', 'n (layers)', 'g', 'q (mmol/g)'],
    'dodo': ['ns (mmol/g)', 'kf', 'nμ (mmol/g)', 'kμ', 'm'],
    'bet': ['n (mmol/g)', 'c']
}

# Default bounds for fitting in the form of (min, max)
# (None, None) represents (-inf, inf)

_MODEL_BOUNDS = {
    'dsl': {
        'q1': (0, None),
        'q2': (0, None),
        'b1': (0, None),
        'b2': (0, None)
    },
    'langmuir': {
        'q': (0, None),
        'b': (0, None)
    },
    'langmuir linear 1': {
        'q': (0, None),
        'b': (0, None)
    },
    'langmuir linear 2': {
        'q': (0, None),
        'b': (0, None)
    },
    'langmuir td': {
        'q': (0, None),
        'b0': (0, None),
        'h': (None, None)
    },
    'gab': {
        'n': (0, None),
        'ka': (0, None),
        'ca': (0, None)
    },
    'mdr': {
        'n0': (0, None),
        'n1': (0, None),
        'a': (0, None),
        'c': (0, None)
    },
    'sips': {
        'q': (0, None),
        'b': (0, None),
        'n': (0, None)
    },
    'toth': {
        'q': (0, None),
        'b': (0, None),
        't': (0, None)
    },
    'bddt': {
        'c': (0, None),
        'n': (0, None),
        'g': (0, None),
        'q': (0, None)
    },
    'dodo': {
        'ns': (0, None),
        'kf': (0, None),
        'nu': (0, None),
        'ku': (0, None),
        'm': (0, None)
    },
    'bet': {
        'n': (0, None),
        'c': (0, None)
    }
}

# Internal conversion from string to function for model fitting
_MODEL_FUNCTIONS = {
    'langmuir': langmuir1,
    'langmuir linear 1': langmuirlin1,
    'langmuir linear 2': langmuirlin2,
    'langmuir td': langmuirTD,
    'dsl': dsl,
    'gab': gab,
    'mdr': mdr,
    'sips': sips,
    'toth': toth,
    'bddt 2n': bddt1,
    'bddt 2n-1': bddt2,
    'bddt': bddt1,
    'dodo': dodo,
    'bet': bet,
}

# Input any temperature dependent model names here, this is to avoid errors where temperature dependent models cause
# problems
_TEMP_DEP_MODELS = ['langmuir td']

# Information for every individual model fitting (WIP)
# _MODEL_INFO = {
#     'langmuir': "This fits to the single site Langmuir isotherm model",
#     'langmuir linear 1': """This fits to the second linearized form of the
#                             single site Langmuir isotherm model. This creates a plot of 1/q vs. 1/P""",
#     'langmuir linear 2': """This fits to the second linearized form of the
#                             single site Langmuir isotherm model""",
#     'langmuir td': langmuirTD,
#     'dsl': dsl,
#     'dsl nc': dsl,
#     'gab': gab,
#     'mdr': mdr,
#     'mdrtd': mdrtd,
#     'sips': sips,
#     'toth': toth,
#     'bddt 2n': bddt1,
#     'bddt 2n-1': bddt2,
#     'bddt': bddt1,
#     'dodo': dodo,
#     'bet': bet,
#     'toth td': tothTD
# }



"""
Utility functions used by the main IsothermFit class and fitting procedure functions
"""
import logging

import numpy as np
from matplotlib import pyplot as plt
from scipy import stats

from src.pyIsoFit.core.model_dicts import _MODEL_DF_TITLES, _MODEL_PARAM_LISTS, _MODEL_BOUNDS
from src.pyIsoFit.core.model_equations import r

logger = logging.getLogger('pyIsoFit-master')

colours = ['b', 'g', 'r', 'c', 'm', 'y', 'tab:orange', 'tab:purple', 'tab:brown', 'tab:olive']

_models_with_b = ['langmuir', 'langmuir linear 1', 'langmuir linear 2', 'sips', 'toth']


def get_xy(df, key_pressures, key_uptakes, model, rel_pres):
    """
    Function that reads dataframe with respect to provided keys and converts it into x and y variables which are
    lists of the x and y co-ordinates for each dataset. This function also converts the x and y axis depending on
    the model used - for 'langmuir linear 1': 1/q vs. 1.p, for 'langmuir linear 2': p/q vs. p, for models
    which use relative pressure: q vs. rel. p

    :param df: pd.DataFrame
                Pure-component isotherm data as a pandas dataframe - must be uptake in mmol/g and pressure in bar or
                equivalent. If datasets at different temperatures are required for fitting, the user must specify
                them in the same dataframe.

    :param key_pressures: list[str]
                List of unique column key(s) which correspond to each dataset's pressure values within the
                dataframe. Can input any number of keys corresponding to any number of datasets in the dataframe.
                If multiple dataframes are specified, make sure keys are identical across each dataframe for each
                temperature. Must be inputted in the same order as key_uptakes and temps.

    :param key_uptakes: list[str]
                List of unique column key(s) which correspond to each dataset's uptake values within the
                dataframe. Can input any number of keys corresponding to any number of datasets in the dataframe.
                If multiple dataframes are specified, make sure keys are identical across each dataframe for each
                temperature. Must be inputted in the same order as key_pressures and temps.

    :param model: str
                Model to be fit to dataset(s).

    :param rel_pres : bool
                Input whether to fit the x axis data to relative pressure instead of absolute

    :return:
        Lists of x and y co-ordinates corresponding to each dataset
    """

    # Importing data and allocating to variables
    x2 = [df[key_pressures[i]].values for i in range(len(key_pressures))]
    y2 = [df[key_uptakes[i]].values for i in range(len(key_pressures))]

    # nan filter for datasets - fixes bug where pandas reads empty cells as nan values
    x_filtr = np.array([np.array(x2[i][~np.isnan(x2[i])]) for i in range(len(x2))])
    y_filtr = np.array([np.array(y2[i][~np.isnan(y2[i])]) for i in range(len(y2))])

    x2 = x_filtr
    y2 = y_filtr

    x = []
    y = []

    # Formatting the x and y axis depending on the model
    if model == "langmuir linear 1":
        for i in range(len(key_pressures)):
            x.append(np.array([1 / p for p in x2[i]]))
            y.append(np.array([1 / q for q in y2[i]]))

    elif model == "langmuir linear 2":
        for i in range(len(key_pressures)):
            y.append(np.array([p / q for (p, q) in zip(x2[i], y2[i])]))
            x.append(np.array(x2[i]))

    elif rel_pres is True:
        for i in range(len(key_pressures)):
            pressure = x2[i]
            x.append(np.array([p / pressure[-1] for p in x2[i]]))
            y.append(np.array(y2[i]))
            del pressure
    else:
        for i in range(len(key_pressures)):
            x.append(np.array(x2[i]))
            y.append(np.array(y2[i]))

    del x2, y2

    return x, y

def plot_settings(log, model=None, rel_pres=False):
    """
    Settings for plotting.

    :param log: tuple(bool)
        tuple input of whether to set a log axis for x and/or y in the shape (x,y)

    :param model: str
        Model to be fit to dataset(s).

    :param rel_pres : bool
        Input whether to fit the x axis data to relative pressure instead of absolute

    Formatting such as axis labels, log plot toggle and param tick settings
    """
    if model == "langmuir linear 1":
        xtitle = '1/Pressure [1/bar]'
        ytitle = '1/Uptake [g/mmol]'

    elif model == "langmuir linear 2":
        xtitle = 'Pressure [bar]'
        ytitle = 'Pressure/uptake [(bar mmol)/g]'

    elif rel_pres:
        xtitle = 'Relative pressure [P/P]'
        ytitle = 'Uptake [mmol/g]'
    else:
        xtitle = "Pressure [bar]"
        ytitle = "Uptake [mmol/g]"

    tick_style = {'direction': 'in',
                  'length': 4,
                  'width': 0.7,
                  'colors': 'black'}

    if log is False:
        log = (False, False)

    if log is True:
        log = (True, True)

    logx = log[0]
    logy = log[1]

    plt.figure(figsize=(8, 6))
    if logx:
        plt.xscale("log")

    if logy:
        plt.yscale("log")

    plt.xlabel(xtitle)
    plt.ylabel(ytitle)
    plt.tick_params(**tick_style)
    plt.grid()


def get_subplot_size(lenx, i):
    """
    Gets the subplot size as a tuple depending on the number of datasets passed.
    This is used for creating henry regime subplots.

    :param lenx: int
        Number of datasets

    :param i:
        Iterator

    :return: tuple of shape (x, y, z)
    """
    if lenx <= 4:
        return 2, 2, i + 1
    if 9 >= lenx > 4:
        return 3, 3, i + 1
    if 16 >= lenx > 9:
        return 4, 4, i + 1


def get_sorted_results(values_dict, model, temps):
    """
    Sorts the values dictionary returned by the generic fitting function from the form:
    {dataset1 : {par_1: value, par_2: value ... par_n: value} ... }

    to the form:
    {par_1: [value(dataset1), value(dataset2), ... value(datasetn)], par_2: ... }

    :param values_dict: dict
        Dictionary of parameters for each dataset to be sorted.

    :param model: str
        Model name for accessing the dictionaries

    """
    # Remove intermediate constraint fitting parameter 'delta'
    for i in range(len(values_dict)):
        if 'delta' in values_dict[i].keys():
            values_dict[i].pop('delta')

    # Create a final result dictionary with the first entry being the the temperatures
    final_results_dict = {'T (K)': temps}

    # Get list of keys for the parameters in the model and create a list of empty lists
    # The number of empty lists is equal to the number of parameters in the model
    param_keys = _MODEL_PARAM_LISTS[model]
    params_list = [[] for _ in range(len(param_keys))]
    for i in range(len(values_dict)):
        for j in range(len(param_keys)):
            params_list[j].append(values_dict[i][param_keys[j]])

    # Get list of titles for the columns within the dataframe - same as _MODEL_PARAM_LISTS but with units
    df_keys = _MODEL_DF_TITLES[model]

    # Allocate the dataframe keys to the fitting results lists and add this to the final results dictionary
    for i in range(len(df_keys)):
        final_results_dict[df_keys[i]] = params_list[i]

    # For error calculations a list of fitting parameters is required for each dataset
    c_list = []
    for i in range(len(values_dict)):
        c_list.append(list(values_dict[i].values()))

    return final_results_dict, c_list


def heat_calc(model, temps, param_dict, x):
    """
    Calculates heat of adsorption for the GAB, Langmuir and DSL isotherm models using the
    temperature dependent parameters.

    For Langmuir and DSL - Van't Hoff:  ln b vs. 1/T
    For GAB - Clausius–Clapeyron approach: ln(x) vs. 1/T
         Plots of 𝑙𝑛(x) vs 1/𝑇 for x at n(x)T1 = n(x)T2 … = n(x)T𝑛

    :param model: str
        Model to be fit to dataset(s).

    :param temps: list[float]
        List of temperatures for calculation of heat of adsorption - must be in K

    :param param_dict: dict
        Dictionary of fitting result lists - this is used for extracting the parameters for plotting


    """
    site = ['A', 'B', 'C']
    t = [1 / i for i in temps]

    def isosteric_heat(t, yparams, j=0):
        ln_b = np.array([np.log(i) for i in yparams])
        mh, bh, rh, ph, sterrh = stats.linregress(t, ln_b)
        h = -0.001 * mh * r

        print("_______________________________________________________________________")
        print("Heat of adsorption for site " + site[j] + ":" + str(round(h, 2)) + " kJ/mol. \n" +
              "R sq of Van't Hoff: " + str(round(rh, 4)))

    if model == "gab":
        xT = []
        h = []
        for i in range(len(x[0])):
            xTT = []
            for p in x:
                xTT.append(p[i])
            xT.append(xTT)
        for xi in xT:
            ln_x = [np.log(i) for i in xi]
            m, b, r2h, p, sterr = stats.linregress(t, ln_x)
            h.append(-0.001 * m * r)

        avgh = np.average(h)
        sterr = (1.96 * np.std(h) / (len(h)) ** 0.5)
        print("ΔH_is: " + str(round(avgh, 2)) + " (∓ " + str(round(sterr, 2)) + ") kJ/mol. ")

    elif model in _models_with_b:
        yparams = param_dict['b (1/bar)']
        isosteric_heat(t, yparams)

    elif model == "dsl":
        yparams = [param_dict['b1 (1/bar)'], param_dict['b2 (1/bar)']]
        for i in range(2):
            isosteric_heat(t, yparams[i], i)
    else:
        return None


def bounds_check(model, cust_bounds, num_comps):
    """
    This function checks whether custom_bounds are inputted into it. If custom_bounds = None, it creates a dictionary
    of default bounds (based on the _MODEL_BOUNDS dictionary) where the key for each parameter is matched with a list
    of tuples which has the same length as the number of datasets being fit.

    If custom bounds are inputted, the above condition is ignored and these are used for fitting

    :param model: str
        Dictionary of parameters for each dataset to be sorted.

    :param cust_bounds: list[tuple[float]]
        List of tuples which include min and max bounds for fitting shape (min, max)

    :param num_comps: int
        Number of components to duplicate the default bounds by

    """
    if cust_bounds is None:
        # Converting bounds to list of tuples corresponding to each dataset
        param_list = _MODEL_PARAM_LISTS[model]
        bounds_dict = _MODEL_BOUNDS[model]
        bounds = {}
        for par in param_list:
            par_list = [bounds_dict[par] for _ in range(num_comps)]
            bounds[par] = par_list
        return bounds
    else:
        bounds = cust_bounds
        return bounds

def save_func(directory, fit_name, filetype, df, comp=''):
    """
    Function that saves fitting results to directory

    :param directory: str
        input directory for the location of the save file

    :param fit_name: str
        input name of file to be saved

    :param filetype: str
        input filetype to be saved

    :param df: pd.DataFrame
        dataframe to be converted into .csv or .json

    :param comp: str
        Used to differentiate files when multiple component fitting results are saved

    """
    fit_string = directory + fit_name + comp + filetype

    file_conv_fit = {
        '.csv': df.to_csv(fit_string),
        '.json': df.to_json(fit_string)
    }
    print("File saved to directory")
    return file_conv_fit[filetype]



import numpy as np
from src.pyIsoFit.core.model_equations import r


def ext_dsl(param_dict, temps, x, comps, yfracs):
    """
    Extended DSL fitting procedure

    :param param_dict: dict
        Dictionary of parameters for co-adsorption calculations

    :param temps: list[float]
        List of temperatures corresponding to each dataset

    :param x: dict[list[list[float]]]
        Dictionary of component datasets' pressure values

    :param comps: list[str]
        List of strings corresponding to the component names

    :param yfracs: list[float]
        List of mole fractions corresponding to each component in the mixture (same order as components)

    :return:
        dictionary of uptake values for each dataset for each component

    Note:
    --------
    First input data is sorted and b1 and b2 are calculated from the heats of adsorption and temperatures.

    Once input data has been sorted numerators of the two terms in the extended dsl equation
    can be worked out. The aim here is to simplify the extended dsl equation into the following form:

    q* = num(siteA) * x / (1 + den(siteA) * x ) + num(siteB) * x / (1 + den(siteB) * x )

    The numerator of site A and B is sorted into a dictionary for its corresponding component.

    The denominators are then calculated - The workaround for this was to
    create a list of the same length as there are the number of temperatures. Within this list
    there are inner lists of the same length as the number of components. This way, we are allocating
    parameters in the following way for example:

    b1 (at 20 C) = 4.2 for component A, 2.4 for component B ... x for component X

    we also do this for b2

    Finally we iterate over the list of lists mentioned previously with the goal of getting:
    (b1A * yA) + (b1B * yB) ... (b1X * yX)

    relating to the denominators of the extended dsl equation:

    1 + (b1A * yA * P) + (b1B * yB * P) = 1 + (b1A * yA + b1B * yB) * P
    """

    def vant_hoff(b0, h, t):
        """
        Inner function for calculating temperature dependent parameter b
        """
        return b0 * np.exp(-h / (r * t))

    def calc_q(x, num_a, num_b, den_a, den_b):
        """

        :param x: list
            pressure values

        :param num_a: float
            numerator for site A

        :param num_b: float
            numerator for site B

        :param den_a: float
            denominator for site A

        :param den_b: float
            denominator for site B

        :return: list
            co-adsorption uptake
        """
        e1 = num_a * x / (1 + den_a * x)
        e2 = num_b * x / (1 + den_b * x)
        return e1 + e2

    params_sorted = {}

    # Step 1 - Calculating b1 and b2 and creating a dictionary of parameter lists for each component
    for i in range(len(comps)):
        params = param_dict[comps[i]]
        x_comp = x[i]
        b1_list_in = []
        b2_list_in = []

        for j in range(len(x_comp)):
            b1_list_in.append(vant_hoff(params['b01'][j], params['h1'][j], temps[j]))
            b2_list_in.append(vant_hoff(params['b02'][j], params['h2'][j], temps[j]))

        params_sorted[comps[i]] = {'q1': params['q1'],
                                   'q2': params['q2'],
                                   'b1': b1_list_in,
                                   'b2': b2_list_in,
                                   'y': yfracs[i]}

    # Step 2- Creating a dictionary of numerators (for site a and b) for each component
    num_dict = {}
    for i in range(len(comps)):
        params = params_sorted[comps[i]]
        num_a_in = []
        num_b_in = []
        for j in range(len(temps)):
            num_a_in.append(params['q1'][j] * params['b1'][j] * params['y'])
            num_b_in.append(params['q2'][j] * params['b2'][j] * params['y'])

        num_dict[comps[i]] = {'a': num_a_in, 'b': num_b_in}

    # Creating a list of b1 and b2 parameters in the form
    # b1_list = [[b1(comp1)(temp1), b1(comp2)(temp1)], [b1(comp1)(temp2), b1(comp2)(temp2)] ... ]
    # The inner list is the same length as the number of components and the outer list is the same length as the
    # number of temperatures.
    b1_list = [[] for _ in range(len(temps))]
    b2_list = [[] for _ in range(len(temps))]
    for i in range(len(temps)):
        for j in range(len(comps)):
            params = params_sorted[comps[j]]
            b1_list[i].append(params['b1'][i])
            b2_list[i].append(params['b2'][i])

    # Calculating denominators and creating a list of denominators
    den_a_list = []
    den_b_list = []
    for i in range(len(temps)):
        b1_in = b1_list[i]
        b2_in = b2_list[i]
        den_a = 0
        den_b = 0
        for j in range(len(b1_in)):
            den_a += b1_in[j] * yfracs[j]
            den_b += b2_in[j] * yfracs[j]
        den_a_list.append(den_a)
        den_b_list.append(den_b)

    # Step 3 - Calculating co-adsorption uptake for each component and creating a dictionary for each component and
    # dataset
    q_dict = {}
    for i in range(len(comps)):
        x_comp = x[i]
        numer = num_dict[comps[i]]
        num_a = numer['a']
        num_b = numer['b']
        q_list = []
        for j in range(len(temps)):
            q = calc_q(np.array(x_comp[j]), num_a[j], num_b[j], den_a_list[j], den_b_list[j])
            q_list.append(q)

        q_dict[comps[i]] = q_list

    return q_dict



"""
Generic fitting procedure for all models except for the constrained dsl fitting procedure
"""
from lmfit import Model, Parameters
from src.pyIsoFit.core.model_fit_def import get_fit_tuples
from src.pyIsoFit.core.model_dicts import _MODEL_FUNCTIONS, _MODEL_PARAM_LISTS


def generic_fit(model, weights, y, guess, temps, cond, meth, cust_bounds, fit_report, henry_constants, henry_off):
    """
    Fits any chosen isotherm model to experimental data and returns the results.

    :param weights:
        Used to perform a weighted fitting - by default this is set equal to self.x within the fitting.py module

    :return: tuple
        Returns a list of ModelResult objects corresponding to each dataframe [0] and a dictionary of parameters created
        by lmfit.
    """
    # Get function from model name string
    isotherm = _MODEL_FUNCTIONS[model]
    # Create instance of class Model with the chosen isotherm model function
    gmod = Model(isotherm)

    params = []
    values_dict = {}
    q_fix = None
    for i in range(len(weights)):
        # Create instance of class Parameters and create parameters using the get_fit_tuples function
        pars = Parameters()
        model_params = get_fit_tuples(model, guess, temps, i, cond, cust_bounds, henry_constants, henry_off, q_fix)
        pars.add_many(*model_params)
        results = gmod.fit(y[i], pars, x=weights[i], method=meth)

        params.append(results)
        values_dict[i] = results.values
        # Optional - call print fitting report generated by lmfit
        if fit_report:
            print("\n\n\n---- FIT REPORT FOR DATASET AT {temp} K -----".format(temp=temps[i]))
            print(results.fit_report())

        # If it is the first iteration and the model has a 'q' parameter, allocate q_fix a value.
        # This results in a float being passed into the q_fix argument of the get_fit_tuples function
        # resulting in a constrained fitting if the fitting condition is set to True
        if i == 0 and 'q' in _MODEL_PARAM_LISTS[model]:
            q_fix = results.values['q']
        elif i == 0 and 'n0' in _MODEL_PARAM_LISTS[model]:
            q_fix = results.values['n0']

        del results, pars

    return params, values_dict
